from inspect import signature, Parameter
from itertools import chain, product, permutations
from typing import Union, Callable, get_type_hints, Any, Tuple, Collection, TypeVar

Self = TypeVar('Self')


def to_type_iter(t: Union[type, None], self_type):
    """
    Convert a type hint to an iteration of concrete types
    :param t: the type hint
    :param self_type: the type to substitute when encountering Self
    :return: a tuple of concrete types
    can handle:
    * types
    * singletons (..., None, Notimplemented)
    * the typing.Any object (equivalent to object)
    * any non-specific typing abstract class (Sized, Iterable, ect...)
    * typing.Union
    """
    if isinstance(t, type):
        return t,
    if t is Any:
        return to_type_iter(object, self_type)
    if t is Self:
        if self_type is _missing:
            raise ValueError('Self cannot be used as a type hint outside of Implementor')
        return to_type_iter(self_type, self_type)
    if t in (..., NotImplemented, None):
        return to_type_iter(type(t), self_type)
    if getattr(t, '__origin__', None) is Union:
        return tuple(chain.from_iterable(to_type_iter(a, self_type) for a in t.__args__))
    if isinstance(getattr(t, '__origin__', None), type) and not t.__args__:
        return to_type_iter(t.__origin__, self_type)
    raise TypeError(f'type annotation {t} is not a type, give it a default to ignore it from the candidate list')


_missing = object()


class Candidate:
    """
    A class representing a specific implementation for a multi-dispatch
    """

    def __init__(self, types, func: Callable, priority):
        """
        :param types: the types for the conditions
        :param func: the function of the implementation
        :param priority: the priority of the candidate over other candidates (higher is tried first)
        """
        self.types = types
        self.func = func
        self.priority = priority
        self.__name__ = getattr(func, '__name__', None)
        self.__doc__ = getattr(func, '__doc__', None)

    @classmethod
    def from_func(cls, priority, func, fallback_type_hint=_missing, self_type=_missing):
        """
        create a list of candidates from function using the function's type hints. ignores all parameters with default
        values, as well as variadic parameters or keyword-only parameters

        :param priority: the priority of the candidate
        :param func: the function to use
        :param fallback_type_hint: the default type hint to use for parameters with missing hints
         this function

        :return: a list of candidates generated from the function
        """
        type_hints = get_type_hints(func)
        params = signature(func).parameters
        type_iters = []
        p: Parameter
        for p in params.values():
            if p.kind not in (p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD) \
                    or p.default is not p.empty:
                break
            t = type_hints.get(p.name, fallback_type_hint)
            if t is _missing:
                raise KeyError(p.name)
            i = to_type_iter(t, self_type)
            type_iters.append(i)
        type_lists = product(*type_iters)
        return [cls(tuple(types), func, priority) for types in type_lists]

    def __str__(self):
        return (self.__name__ or 'unnamed candidate') + '<' + ', '.join(n.__name__ for n in self.types) + '>'

    def permutations(self, first_priority_offset=0.5):
        """
        create a list of candidates from a single candidate, such that they all permutations
        of the candidate's types will be accepted. Useful for symmetric functions.

        :param first_priority_offset: the priority increase to give the first output of the list,
         to avoid priority collisions

        :return: a list of equivalent candidates, differing only by the type order
        """
        if not self.types:
            raise ValueError("can't get permutations of a 0-parameter candidate")
        ret = []
        name = getattr(self.func, '__name__', None)
        call_args = ', '.join('_' + str(i) for i in range(len(self.types)))
        seen = set()
        glob = {'__original__': self.func}
        first = True
        for perm in permutations(range(len(self.types))):
            if first:
                func = self.func
                t = self.types
                if t in seen:
                    continue
                seen.add(t)
                priority = self.priority + first_priority_offset
                first = False
            else:
                t = tuple(self.types[i] for i in perm)
                if t in seen:
                    continue
                seen.add(t)
                args = ", ".join('_' + str(i) for i in perm)
                ns = {}
                exec(
                    f"def func({args}): return __original__({call_args})",
                    glob, ns
                )
                func = ns['func']
                if name:
                    func.__name__ = name
                priority = self.priority
            ret.append(
                Candidate(t, func, priority)
            )
        return ret


def cmp_key(rhs: Tuple[type, ...], lhs: Tuple[type, ...]):
    """
    check whether two type tuples are ordered in any way
    :return: -1 if rhs is a sub-key of lhs, 1 if lhs is a sub-key of rhs, 0 if the two keys cannot be compared
     (it is an error to sent identical keys)
    """
    ret = 0
    for r, l in zip(rhs, lhs):
        if r is l:
            continue
        elif issubclass(r, l):
            if ret == 1:
                return 0
            ret = -1
        elif issubclass(l, r):
            if ret == -1:
                return 0
            ret = 1
        else:
            return 0
    return ret


def get_least_key_index(candidates: Collection[Candidate]):
    """
    :param candidates: a collection of candidates
    :return: the index of the candidate with a least-key from among the candidates, or -1 if none exists
    """
    # todo test
    if len(candidates) < 2:
        return -1
    i = iter(candidates)
    ret = next(i)
    ret_ind = 0
    for ind, k in enumerate(i, 1):
        cmp = cmp_key(ret.types, k.types)
        if cmp == 0:
            return -1
        if cmp == 1:
            ret_ind = ind
            ret = k
    return ret_ind



from __future__ import annotations

from functools import partial
from itertools import chain
from numbers import Number
from typing import Dict, Tuple, List, Callable, Union, Optional, Iterable

from sortedcontainers import SortedDict

from dyndis.candidate import Candidate, get_least_key_index
from dyndis.descriptors import MultiDispatchOp, MultiDispatchMethod, MultiDispatchStaticMethod
from dyndis.implementor import Implementor
from dyndis.trie import Trie
from dyndis.util import RawReturnValue, AmbiguityError, NoCandidateError

CandTrie = Trie[type, Dict[Number, Candidate]]

RawNotImplemented = RawReturnValue(NotImplemented)


def process_new_layers(layers: Iterable[List[Candidate]]):
    ret = []
    for layer in layers:
        while True:
            least_ind = get_least_key_index(layer)
            if least_ind == -1:
                break
            least = layer.pop(least_ind)
            ret.append((least,))
        ret.append(layer)
    return ret


def add_priority(seen_priorities: SortedDict[Number, List[Candidate]], candidate: Candidate):
    """
    Add a candidate to a temporary result dictionary
    :param seen_priorities: the sorted dict of existing candidates
    :param candidate: the candidate to add
    """
    eq_prio_list = seen_priorities.get(candidate.priority)
    if not eq_prio_list:
        eq_prio_list = seen_priorities[candidate.priority] = []
    eq_prio_list.append(candidate)


class CachedSearch:
    """
    A cached search for candidates of a specific type tuple
    """

    def __init__(self, owner: MultiDispatch, key: Tuple[type, ...]):
        """
        :param owner: the owning multidispatch
        :param key: the type tuple to use
        """
        ex = owner.candidates.get(key)
        if ex:
            lay_0 = [[v] for v in reversed(ex.values())]
        else:
            lay_0 = []

        self.sorted: List[List[Candidate]] = lay_0
        self.next_layer: List[Tuple[int, CandTrie]] = [(0, owner.candidates)]
        self.err: Optional[Exception] = None
        self.query = key

    def __iter__(self):
        yield from self.sorted
        while self.next_layer:
            yield from self.advance()

    def advance(self) -> Iterable[List[Candidate]]:
        """
        advance the search into the next layer, caching the result

        :return: the newly added candidates of this layer
        """
        ret = SortedDict()
        nexts: List[Tuple[int, CandTrie]] = []
        for (depth, n_trie) in self.next_layer:
            self.advance_search_inexact(n_trie, depth, ret, nexts)

        self.next_layer = nexts
        new_layers = process_new_layers(reversed(ret.values()))
        self.sorted.extend(new_layers)
        return new_layers

    def advance_search_inexact(self, current_trie: CandTrie, current_depth: int,
                               results: SortedDict[Number, List[Candidate]], nexts: List[Tuple[int, CandTrie]]):
        """
        advance the search by looking for candidates with at least one mismatch

        :param current_trie: the trie to conduct the search in
        :param current_depth: the current depth of the trie
        :param results: a list of candidates that match
        :param nexts: a list of tries to search in for the next layer
        """
        if current_depth == len(self.query):
            return
        curr_key = self.query[current_depth]
        for child_type, child in current_trie.children.items():
            if child_type is curr_key:
                self.advance_search_inexact(child, current_depth + 1, results, nexts)
            elif issubclass(curr_key, child_type):
                self.advance_search(child, current_depth + 1, results, nexts)

    def advance_search(self, current_trie: CandTrie, current_depth: int,
                       results: SortedDict[Number, List[Candidate]], nexts: List[Tuple[int, CandTrie]]):
        """
        advance the search by looking for candidates without any mismatches

        :param current_trie: the trie to conduct the search in
        :param current_depth: the current depth of the trie
        :param results: a list of candidates that match
        :param nexts: a list of tries to search in for the next layer
        """
        if current_depth == len(self.query):
            curr_value = current_trie.value(None)
            if curr_value:
                for candidate in curr_value.values():
                    add_priority(results, candidate)
            return
        curr_key = self.query[current_depth]
        next_exact = current_trie.children.get(curr_key)
        if next_exact:
            self.advance_search(next_exact, current_depth + 1, results, nexts)
        nexts.append((current_depth, current_trie))


EMPTY = object()


class MultiDispatch:
    """
    The central class, a callable that can delegate to multiple candidates depending on the types of parameters
    """

    def __init__(self, name: str = None, doc: str = None):
        """
        :param name: an optional name for the callable
        :param doc: an optional doc for the callable
        """
        self.__name__ = name
        self.__doc__ = doc

        self.candidates: CandTrie = Trie()
        self.cache: Dict[int, Dict[Tuple[type, ...], CachedSearch]] = {}

    def _clean_cache(self, sizes: Iterable[int]):
        """
        clear the candidate cache for all type tuples of the sizes specified

        :param sizes: the sizes for which to clear to cache
        """
        for size in sizes:
            self.cache.pop(size, None)

    def _add_candidate(self, candidate: Candidate, clean_cache=True):
        """
        Add a single candidate to the multidispatch. If the multidispatch has no set name or doc, the name or doc of
        the candidate will be used (if available)

        :param candidate: the candidate to add
        :param clean_cache: whether to clean the relevant cache
        """
        sd = self.candidates.get(candidate.types)
        if sd is None:
            sd = self.candidates[candidate.types] = SortedDict()
        if candidate.priority in sd:
            raise ValueError(f'cannot insert candidate, a candidate of equal types ({candidate.types})'
                             f' and priority ({candidate.priority}) exists ')
        sd[candidate.priority] = candidate

        if not self.__name__:
            self.__name__ = candidate.__name__
        if not self.__doc__:
            self.__doc__ = candidate.__doc__
        if clean_cache:
            self._clean_cache((len(candidate.types),))

    def add_candidates(self, candidates: Iterable[Candidate]):
        """
        Add a collection of candiates to the multidispatch. If the multidispatch has no set name or doc, the name or doc of the first candidate with the relevant attributes will be used.

        :param candidates: an iterable of candidates to be added.
        """
        clean_sizes = set()
        for cand in candidates:
            self._add_candidate(cand, clean_cache=False)
            clean_sizes.add(len(cand.types))
        self._clean_cache(clean_sizes)
        return self

    def add_func(self, priority=0, symmetric=False, func=None):
        """
        Adds candidates to a multidispatch generated from a function, usable as a decorator

        :param priority: the priority of the candidates.
        :param symmetric: if set to true, the permutations of all the candidates are added as well
        :param func: the function to used
        """
        if not func:
            return partial(self.add_func, priority, symmetric)
        cands = Candidate.from_func(priority, func)
        if symmetric:
            cands = chain.from_iterable(c.permutations() for c in cands)
        self.add_candidates(cands)
        return self

    def _yield_candidates(self, types):
        """
        yield all the relevant candidates for a type tuple, sorted first by number of upcasts required (ascending),
        and second by priority (descending)

        :param types: the type tuple to get candidates for
        """
        sub_cache = self.cache.get(len(types))
        if not sub_cache:
            sub_cache = self.cache[len(types)] = {}
            cache = sub_cache[types] = CachedSearch(self, types)
        else:
            cache = sub_cache.get(types)
            if not cache:
                cache = sub_cache[types] = CachedSearch(self, types)

        for layer in cache:
            if len(layer) != 1:
                raise AmbiguityError(layer, types)
            yield layer[0]

    def get(self, args, kwargs, default=None):
        """
        call the multidispatch with args as arguments, attempts all the appropriate candidate until one returns a
        non-NotImplemted value. If all the candidates are exhausted, returns default.

        :param args: the arguments for the multidispatch
        :param kwargs: keyword arguments forwarded directly to any attempted candidate
        :param default: the value to return if all candidates are exhausted
        """
        types = tuple(type(a) for a in args)
        for c in self._yield_candidates(types):
            ret = c.func(*args, **kwargs)
            if ret is not NotImplemented:
                return RawReturnValue.unwrap(ret)
        return default

    def __call__(self, *args, **kwargs):
        """
        call the multidispatch and raise an error if no candidates are found
        """
        ret = self.get(args, kwargs, default=EMPTY)
        if ret is EMPTY:
            raise NoCandidateError(args)
        return ret

    def op(self):
        """
        :return: an adapter for the multidispatch to be used as an adapter, returning NotImplemented if no candidates match,
         and setting the multidispatch's name if necessary
        """
        return MultiDispatchOp(self)

    def method(self):
        """
        :return: an adapter for the multidispatch to be used as a method, raising error if no candidates match,
         and setting the multidispatch's name if necessary
        """
        return MultiDispatchMethod(self)

    def staticmethod(self):
        """
        :return: an adapter for the multidispatch to be used as a static method, raising error if no candidates match,
         and setting the multidispatch's name if necessary
        """
        return MultiDispatchStaticMethod(self)

    def implementor(self, *args, **kwargs) -> Union[Callable[[Callable], 'Implementor'], 'Implementor']:
        """
        create an Implementor for the MultiDispatch and call its implementor method with the arguments
        """
        return Implementor(self).implementor(*args, **kwargs)

    def __str__(self):
        if self.__name__:
            return f'<MultiDispatch {self.__name__}>'
        return super().__str__()



from __future__ import annotations

from types import MappingProxyType
from typing import TypeVar, Generic, Dict, Tuple, Iterable, Iterator, MutableMapping, List, Callable, Any, Optional

K = TypeVar('K')
V = TypeVar('V')

_missing = object()
_no_default = object()


class Trie(Generic[K, V], MutableMapping[Iterable[K], V]):
    """
    A generic non-compressing trie that supports manual search. Acts as a mapping of iterables to keys
    """
    _empty_val: V
    _len: int

    def __init__(self, map=(), **additionals):
        self._children: Dict[K, Trie[K, V]] = {}
        self.children = MappingProxyType(self._children)

        self.clear()

        self.update(map, **additionals)

    def _set_default_child(self, k):
        """
        ensure the trie has a child for key k, and returns it
        :param k: the key to ensure a child exists for
        :return: the child for key k
        """
        ret = self._children.get(k, None)
        if ret is None:
            ret = self._children[k] = type(self)()
        return ret

    def _set(self, i: Iterator[K], v: V, override: bool):
        """
        Set a value in the trie
        :param i: an iterator of keys for the trie and its children
        :param v: the value to set for the key
        :param override: whether to destroy an existing key
        :return: 2-tuple: the change in lens, the current value for the key
        """
        try:
            n = next(i)
        except StopIteration:
            if self._empty_val is _missing:
                delta = 1
                val = self._empty_val = v
                self._len += 1
            else:
                delta = 0
                if override:
                    val = self._empty_val = v
                else:
                    val = self._empty_val
            return delta, val
        else:
            child = self._set_default_child(n)
            delta, val = child._set(i, v, override)
            self._len += delta
            return delta, val

    def _get(self, i: Iterator[K], d: V):
        """
        Get the value in a trie
        :param i: an iterator of keys for the trie and its children
        :param d: the default value to return if the value does not exist
        :return: the value of the trie of the final key, or d if none exists
        """
        try:
            n = next(i)
        except StopIteration:
            return self.value(d)
        else:
            child = self._children.get(n)
            if not child:
                return d
            return child._get(i, d)

    def _pop(self, i: Iterator[K]):
        """
        Remove a key and its value from the trie, will delete empty child tries
        :param i: an iterator of keys for the trie and its children
        :return: a 3-tuple: whether an item was found,
         the child key the item was found in if it empty,
         and the value of the item
        """
        try:
            n = next(i)
        except StopIteration:
            if self._empty_val is _missing:
                return False, None, None
            else:
                self._len -= 1
                popped = self._empty_val
                self._empty_val = _missing
                return True, _missing, popped
        else:
            child = self._children.get(n)
            if not child:
                return False, None, None
            success, _, popped = child._pop(i)
            if not success:
                return False, None, None
            self._len -= 1
            if self._len == 0:
                # don't bother dropping children, since you're gonna get dropped anyway, return the key needed to drop
                return True, n, popped

            if len(child) == 0:
                del self._children[n]

            return True, _missing, popped

    def _pop_item(self, buffer: List[K]):
        """
        Remove a key-value pair from the trie via DFS, will delete empty child tries
        :param buffer: a list containing the complete key of the deleted pair
        :return: the value deleted
        """
        if self._empty_val is not _missing:
            ret = self._empty_val
            self._empty_val = _missing
            self._len -= 1
            return ret

        t: Trie
        k, t = next(iter(self._children.items()))
        buffer.append(k)
        ret = t._pop_item(buffer)
        self._len -= 1
        if not t:
            if self or not buffer:
                del self._children[k]
        return ret

    def _items(self):
        """
        Iterate over all key-value pairs in the trie an its children (using DFS)
        Warning: each is returned using the same key buffer, convert to list or other sequency type
         prior to returning to user
        """
        buffer = []
        if self._empty_val is not _missing:
            yield buffer, self._empty_val
        stack: List[Iterator[Tuple[K, Trie[K, V]]]] = [iter(self._children.items())]
        while stack:
            last_iter = stack[-1]
            try:
                k, t = next(last_iter)
            except StopIteration:
                stack.pop()
                if not stack:
                    return
                buffer.pop()
                continue
            else:
                buffer.append(k)
                if t._empty_val is not _missing:
                    yield buffer, t._empty_val
                stack.append(iter(t._children.items()))

    def __setitem__(self, string: Iterable[K], value: V):
        self._set(iter(string), value, True)

    def __getitem__(self, item):
        ret = self._get(iter(item), _missing)
        if ret is _missing:
            raise KeyError(item)
        return ret

    def __len__(self):
        return self._len

    def __delitem__(self, key):
        self.pop(key)

    def __iter__(self):
        return self.keys(tuple)

    def __contains__(self, item):
        return self._get(iter(item), _missing) is not _missing

    def keys(self, key_joiner: Optional[Callable[[List], Any]] = tuple):
        """
        get an iterable of keys
        :param key_joiner: an optional function to convert the key-buffer to permanent data (default is tuple)
        """
        if key_joiner:
            return (key_joiner(k) for k, _ in self._items())
        else:
            return (k for k, _ in self._items())

    def items(self, key_joiner: Optional[Callable[[List], Any]] = tuple):
        """
        get an iterable of key-value pairs
        :param key_joiner: an optional function to convert the key-buffer to permanent data (default is tuple)
        """

        if key_joiner:
            return ((key_joiner(k), v) for k, v in self._items())
        else:
            return ((k, v) for k, v in self._items())

    def values(self):
        if self._empty_val is not _missing:
            yield self._empty_val
        for c in self._children.values():
            yield from c.values()

    def get(self, k, default=None)->V:
        return self._get(iter(k), default)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return len(self) == len(other) \
               and self._empty_val == other._empty_val \
               and self._children == other._children

    def pop(self, key, default=_no_default):
        success, child_key, ret = self._pop(iter(key))
        if not success:
            if default is _no_default:
                raise KeyError(key)
            return default
        if child_key is not _missing:
            # a 0-length child needs dropping
            assert len(self._children[child_key]) == 0
            del self._children[child_key]
        return ret

    def popitem(self, key_joiner: Optional[Callable[[List], Any]] = tuple):
        if not self:
            raise KeyError
        buffer = []
        v = self._pop_item(buffer)
        if buffer:
            first = buffer[0]
            if not self.children[first]:
                del self.children[first]
        if key_joiner not in (None, list):
            key = key_joiner(buffer)
        else:
            key = buffer
        return key, v

    def clear(self):
        self._len = 0
        self._empty_val = _missing
        self._children.clear()

    def setdefault(self, k, default=None):
        return self._set(iter(k), default, override=False)[1]

    def value(self, default=_no_default):
        """
        Get the current internal value of the trie, equivelant to `self[()]`
        :param default: the optional value to return if a the trie has no value, by default, will raise KeyError
        """
        ret = self._empty_val
        if ret is _missing:
            if default is _no_default:
                raise KeyError()
            return default
        return ret

    def has_value(self):
        """
        :return: whether the trie has in internal value, equivelant to `() in self`
        """
        return self._empty_val is not _missing



from dyndis.multidispatch import MultiDispatch, RawNotImplemented, RawReturnValue
from dyndis.candidate import Self
from dyndis.util import AmbiguityError
from dyndis_data import __version__, __author__, __url__



from typing import Any, NamedTuple


class RawReturnValue(NamedTuple):
    """
    A class to wrap otherwise special return values from a multidispatch candidate
    """
    inner: Any

    @classmethod
    def unwrap(cls, x):
        """
        If x is a RawReturnValue, return its inner value, if not, return x unchanged
        """
        if isinstance(x, cls):
            return x.inner
        return x


class AmbiguityError(RuntimeError):
    """An error indicating that a multidispatch had to decide between candidates of equal precedence"""

    def __init__(self, candidates, types):
        cands = "[" + ", ".join(str(c) for c in candidates) + "]"
        super().__init__(f'multiple candidates of equal precedence: {cands} for key <' + ", ".join(t.__name__ for t in types) + ">")


class NoCandidateError(TypeError):
    """An error indicating that a multidispatch has no applicable candidates"""

    def __init__(self, args):
        super().__init__('no valid candidates for argument types <' + ", ".join(type(a).__name__ for a in args) + '>')



import setuptools

import dyndis_data as dyndis

setuptools.setup(
    name='dyndis',
    version=dyndis.__version__,
    url=dyndis.__url__,
    author=dyndis.__author__,
    packages=['dyndis', 'dyndis_data'],
    python_requires='>=3.7.0',
    include_package_data=True,
    data_files=[
        ('', ['README.md', 'CHANGELOG.md']),
    ],
    install_requires=["sortedcontainers"],
    extras_require={
        "testing": ['numpy']
    }
)
